#define SCROLLTIME 6 // loop count value, the smaller is faster (fastest: 1)

// hw defines:
#define COMMON_CATHODE       //common cathode rows, common anode columns !!!

#define DATA_PIN  RB3_bit //PORTB.3, the hc595 data

//Store pin: PORTA.3  (RA3), positive clock for shiftregs output storage
#define StorePulse RA3_bit = 1; RA3_bit = 0

//clock pin: PORTB.0  (RB0), positive clock for shift in
#define ClockPulse RB0_bit = 1; RB0_bit = 0

#define SCROLL_PIN RB4_bit   // a jumper for (re)start (open) / stop (closed) scrolling

//Line decoder inputs: PORTA<2:0> ;RA0 = LSB
#define DisableLines PORTA = 0b0111  //= 7, all 1 (the outputs are active low)

#include "bitmap7.h" // ASCII font bitmap table, 7 bytes/char

const char message[] = "HELP US"; // 0 = end sign

//global arrays
char pattbuf[7];   // holds the actual char's line patterns
char linebuf[7*7]; // 7 bytes for the first (uppermost) line (row), 7 bytes for the 2nd, ...

void HW_Init(void) {
    ADCON1=0x07;
    CMCON = 0b00000111; // switch off comparators
    DisableLines;       // PORTA = 7, also store pin is zeroed
    PORTB = 0;          // also data and clock pins are zeroed
    TRISA = 0;          // PORTA pins are outputs
    TRISB = 0xF6;       // only PORTB.0,2,3 are outputs
    NOT_RBPU_bit = 0;   // portb pullups on

/* init timer2 for line timing
    one row display = 870 us (without timer2 waiting) but
    with scroll (worst case) (10+48+435+1656 = 2.149 ms)
    sum = 2149 + 870 = 3019 us (min. for timer2)

    Prescaler=1:1; PostScaler=1:12; Period = 3.024 ms
    4 MHz: 3.028 ms/line, 21.35 ms the whole loop (46.8 Hz refresh rate)
*/
    PR2 = 252;           // Timer2 Match value
    T2CON = 0b01011100;  // enable timer2 with 1:12 postscaler, 1:1 prescaler
    TMR2IF_bit = 0;      // clear IT flag
    PIE1 = 2; // only TMR2IE_bit = 1; enable Timer2 IT

    INTCON = 0xC0; // Enable global (GIE_bit) and periferial (PEIE_bit) ITs
}

void interrupt(void) {
    DisableLines;   // PORTA = 7
    TMR2ON_bit = 0; // disable timer2
    TMR2IF_bit = 0; // clear IT flag
}

void get_patterns(char ch) {
     unsigned bitmap_idx;
     char rowcnt;

     ch -= 32; // the space is the first char in the bitmap table
     //calculate the bitmap table offset of the first pattern of the char ch
     bitmap_idx = 8*ch - ch; // = 7*ch, but faster

     rowcnt = 0; //we have 7 patterns (0...6)
     do {
        ch = bitmap[bitmap_idx + rowcnt]; // read pattern from the bitmap table
        ch <<= 2; // drop unused two bits (7,6)
#ifdef COMMON_CATHODE
        ch = ~ch; // inverted (0<>1)
#endif
        pattbuf[rowcnt] = ch; // and store it
     } while (++rowcnt != 7);
}

char nextchar(void) {
     char ch;
     static unsigned msg_idx = 0; // 16bit var, a very long msg is also possible

     for (;;) {
         if (ch = message[msg_idx++])
            break;    // not 0 (end sign)
         msg_idx = 0; // else back to begin
     }
     return ch;
}

void shift_in_a_column(void) {
     char pat_idx, buf_idx, ch, rowcnt, bytecnt, carry_over;

     pat_idx = 6;  // last (7th, lowermost) pattern
     buf_idx = 48; // last (7th, lowermost) line's last (7th) byte

     rowcnt = 7;
     do {
        // unfortunately (in MikroC) no rotate with carry

        // pattern rotate left with carry
        ch = pattbuf[pat_idx];
        carry_over = ch & 0x80;
        ch <<= 1; // a 0 shifted in from right but not used
        pattbuf[pat_idx--] = ch;

        // rotate a whole line (7 bytes) left with carry
        bytecnt = 7;
        do {
           ch = linebuf[buf_idx];
           ch <<= 1;
           if (carry_over) ch.B0 = 1;
           linebuf[buf_idx--] = ch;
           carry_over = ch & 0x80;
        } while (--bytecnt);
     } while (--rowcnt);
}

void display(void) {
   char ch, buf_idx, bytecnt, bitcnt, rowcnt;

   buf_idx = 0; //from begin of buffer
   // loop until rowcnt = 7
   rowcnt = 0;
   do { // 870 us / row
      bytecnt = 7;
      do {
         ch = linebuf[buf_idx++];
         bitcnt = 8;
         do {
            DATA_PIN = ch.B7;
            ch <<= 1;
            ClockPulse;
         } while (--bitcnt);
      }while (--bytecnt);

      //timed display (same row brightness)
      do {} while (TMR2ON_bit); //wait for timer2 IT (timer2 and all lines off)
      TMR2ON_bit = 1; //restart timer2, also a time gap to switch off drivers

      StorePulse;
      PORTA = rowcnt; // switch on the actual line driver
   } while (++rowcnt != 7);
}

void clear_linebuf(void) {
  char cnt = 49;
  do {
#ifdef COMMON_CATHODE
     linebuf[cnt-1] = 0;
#else
     linebuf[cnt-1] = 255;
#endif
  } while (--cnt);
}

void main(void) {
  char width, scrollcnt, ch;


  HW_Init();
  clear_linebuf();
  //for prompt scroll:
  width = 0;
  scrollcnt = 1;

  for (;;) {
    if (SCROLL_PIN && --scrollcnt == 0) {
       scrollcnt = SCROLLTIME;
       if (!width) {
          ch = nextchar();  //48 us
          width = 6;
          if (ch == 32) width = 3;
          get_patterns(ch); //435 us
       }
       --width;
       shift_in_a_column(); //1656 us !!!!
    }
    display(); //6.08 ms (without timer2 waiting !) 6.08/7 = appr 870 us / row
  }
}
//------ end